==> kafka_tools_abm-2.0.56359a0.tar.gz: kafka_tools_abm-2.0.56359a0/kafka_tools/record_processor/record_processor.py <==
#  Copyright (C) 2022
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasiliev Ivan <i.vasiliev@technokert.ru>
import asyncio
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Callable, Awaitable, Optional, Any
from aiokafka import ConsumerRecord
from async_tools import acall, AsyncOnStart, AsyncOnStop
from .kafka_record import KafkaRecord, PreproceededRecord
from ..abstract_buffer import AbstractBuffer
from ..kafka_client.consumer import Consumer

logger = logging.getLogger(__file__)


class PreProcessRecordError(ValueError):
    pass


class DeserializeRecordError(PreProcessRecordError):
    pass


class AbstractRecordProcessor(ABC):
    @dataclass
    class Config:
        topic: str

    @dataclass
    class Context:
        consumer: Consumer

    def __init__(self, config: Config, context: Context, record_callback: Callable) -> None:
        self.context = context
        self.config = config
        self._record_callback = record_callback
        self.context.consumer.subscribe(self.config.topic, self.process_record)

    @abstractmethod
    async def process_record(self, record: ConsumerRecord) -> None:
        pass

    @abstractmethod
    async def _preprocess_record(self, msg: ConsumerRecord) -> PreproceededRecord:
        pass

    @staticmethod
    def form_record(pre_proceeded_record: PreproceededRecord, consumer_record: ConsumerRecord) -> KafkaRecord:
        return KafkaRecord(key=pre_proceeded_record.key,
                           value=pre_proceeded_record.value,
                           topic=consumer_record.topic,
                           partition=consumer_record.partition,
                           offset=consumer_record.offset,
                           timestamp=consumer_record.timestamp,
                           timestamp_type=consumer_record.timestamp_type,
                           checksum=consumer_record.checksum,
                           serialized_key_size=consumer_record.serialized_key_size,
                           serialized_value_size=consumer_record.serialized_value_size,
                           headers=consumer_record.headers
                           )


class AbstractDeserializedRecordProcessor(AbstractRecordProcessor, ABC):
    def __init__(self,
                 config: AbstractRecordProcessor.Config,
                 context: AbstractRecordProcessor.Context,
                 record_callback: Callable,
                 key_deserializer: Callable[[bytes], Any] | None = None,
                 value_deserializer: Callable[[bytes], Any] | None = None) -> None:
        super().__init__(config, context, record_callback)
        self.context = context
        self.key_deserializer = key_deserializer
        self.value_deserializer = value_deserializer

    async def _preprocess_record(self, record: ConsumerRecord) -> PreproceededRecord:
        try:
            key = record.key
            key = key if key is None or self.key_deserializer is None else self.key_deserializer(key)
            value = record.value
            value = value if value is None or self.value_deserializer is None else self.value_deserializer(value)
            return PreproceededRecord(key=key, value=value)
        except Exception as e:
            raise DeserializeRecordError(e)


class AbstractSingleRecordProcessor(AbstractRecordProcessor, ABC):

    def __init__(self,
                 config: AbstractRecordProcessor.Config,
                 context: AbstractRecordProcessor.Context,
                 record_callback: Callable[[KafkaRecord], Awaitable[None]]) -> None:
        super().__init__(config, context, record_callback)
        self._record_callback = record_callback

    async def process_record(self, record: ConsumerRecord) -> None:
        try:
            pre_proceeded_record = await self._preprocess_record(record)
            formed_record = self.form_record(pre_proceeded_record, record)
            await acall(self._record_callback(formed_record))
        except PreProcessRecordError as e:
            # TODO consider add option to skip preprocessing exception and commit anyway
            logger.warning(f"got error in record preprocessing; record {record}, error: {repr(e)}")
            raise e
        await self.context.consumer.commit(record)


class AbstractBufferProcessor(AbstractRecordProcessor, AsyncOnStart, AsyncOnStop, ABC):
    @dataclass
    class Config(AbstractBuffer.Config, AbstractRecordProcessor.Config):
        exception_amount_threshold = 5

    @dataclass
    class Context(AbstractRecordProcessor.Context):
        pass

    def __init__(self,
                 config: Config,
                 context: Context,
                 record_callback: Callable[[list[KafkaRecord]], Awaitable[None]]) -> None:
        super().__init__(config, context, record_callback)
        self.config = config
        self._buffer = AbstractBuffer(self._handle_buffer, config)
        self._buffer_task: Optional[asyncio.Task] = None

    async def _on_start(self) -> None:
        self._buffer_task = asyncio.create_task(self._buffer.run())

    async def _on_stop(self) -> None:
        self._buffer_task.cancel()

    async def process_record(self, record: ConsumerRecord) -> None:
        await self._buffer.add_to_buffer(record)

    async def _handle_buffer(self, buffer: list[ConsumerRecord]) -> None:
        logger.debug("handle buffer")

        last_record = buffer[-1]
        formed_records = []
        exception_amount = 0
        for record in buffer:
            try:
                pre_proceeded_record = await self._preprocess_record(record)
                formed_records.append(self.form_record(pre_proceeded_record, record))
            except PreProcessRecordError as e:
                # TODO consider add option to skip preprocessing exception and commit anyway
                logger.warning(f"got preprocess error on record {record}, error: {repr(e)}")
                exception_amount += 1
                if exception_amount >= self.config.exception_amount_threshold:
                    # TODO fix; not working properly
                    raise e
        await acall(self._record_callback(formed_records))
        await self.context.consumer.commit(last_record)


class RecordProcessor(AbstractDeserializedRecordProcessor, AbstractSingleRecordProcessor):
    @dataclass
    class Config(AbstractSingleRecordProcessor.Config):
        pass

    @dataclass
    class Context(AbstractDeserializedRecordProcessor.Context):
        pass


class BufferRecordProcessor(AbstractDeserializedRecordProcessor, AbstractBufferProcessor):
    @dataclass
    class Config(AbstractBufferProcessor.Config, AbstractRecordProcessor.Config):
        pass

    @dataclass
    class Context(AbstractDeserializedRecordProcessor.Context):
        pass

    def __init__(self,
                 config: Config,
                 context: Context,
                 record_callback: Callable[[list[KafkaRecord]], Awaitable[None]],
                 key_deserializer: Callable[[bytes], Any] | None = None,
                 value_deserializer: Callable[[bytes], Any] | None = None) -> None:
        super().__init__(config, context, record_callback,
                         key_deserializer=key_deserializer, value_deserializer=value_deserializer)

==> kafka_tools_abm-2.0.56430a0.tar.gz: kafka_tools_abm-2.0.56430a0/kafka_tools/record_processor/record_processor.py <==
#  Copyright (C) 2022
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasiliev Ivan <i.vasiliev@technokert.ru>
import asyncio
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Callable, Awaitable, Optional
from aiokafka import ConsumerRecord
from async_tools import acall, AsyncOnStart, AsyncOnStop
from ..deserializer import RecordDeserializer
from .kafka_record import KafkaRecord, PreproceededRecord
from ..abstract_buffer import AbstractBuffer
from ..kafka_client.consumer import Consumer

logger = logging.getLogger(__file__)


class PreProcessRecordError(ValueError):
    pass


class DeserializeRecordError(PreProcessRecordError):
    pass


class AbstractRecordProcessor(ABC):
    @dataclass
    class Config:
        topic: str

    @dataclass
    class Context:
        consumer: Consumer

    def __init__(self, config: Config, context: Context, record_callback: Callable) -> None:
        self.context = context
        self.config = config
        self._record_callback = record_callback
        self.context.consumer.subscribe(self.config.topic, self.process_record)

    @abstractmethod
    async def process_record(self, record: ConsumerRecord) -> None:
        pass

    @abstractmethod
    async def _preprocess_record(self, msg: ConsumerRecord) -> PreproceededRecord:
        pass

    @staticmethod
    def form_record(pre_proceeded_record: PreproceededRecord, consumer_record: ConsumerRecord) -> KafkaRecord:
        return KafkaRecord(key=pre_proceeded_record.key,
                           value=pre_proceeded_record.value,
                           topic=consumer_record.topic,
                           partition=consumer_record.partition,
                           offset=consumer_record.offset,
                           timestamp=consumer_record.timestamp,
                           timestamp_type=consumer_record.timestamp_type,
                           checksum=consumer_record.checksum,
                           serialized_key_size=consumer_record.serialized_key_size,
                           serialized_value_size=consumer_record.serialized_value_size,
                           headers=consumer_record.headers
                           )


class AbstractDeserializedRecordProcessor(AbstractRecordProcessor, ABC):
    @dataclass
    class Context(AbstractRecordProcessor.Context):
        deserializer: RecordDeserializer

    def __init__(self,
                 config: AbstractRecordProcessor.Config,
                 context: Context,
                 record_callback: Callable) -> None:
        super().__init__(config, context, record_callback)
        self.context = context

    async def _preprocess_record(self, record: ConsumerRecord) -> PreproceededRecord:
        try:
            return self.context.deserializer.deserialize_record(record)
        except Exception as e:
            raise DeserializeRecordError(e)


class AbstractSingleRecordProcessor(AbstractRecordProcessor, ABC):

    def __init__(self, config: AbstractRecordProcessor.Config, context: AbstractRecordProcessor.Context,
                 record_callback: Callable[[KafkaRecord], Awaitable[None]]) -> None:
        super().__init__(config, context, record_callback)
        self._record_callback = record_callback

    async def process_record(self, record: ConsumerRecord) -> None:
        try:
            pre_proceeded_record = await self._preprocess_record(record)
            formed_record = self.form_record(pre_proceeded_record, record)
            await acall(self._record_callback(formed_record))
        except PreProcessRecordError as e:
            # TODO consider add option to skip preprocessing exception and commit anyway
            logger.warning(f"got error in record preprocessing; record {record}, error: {repr(e)}")
            raise e
        await self.context.consumer.commit(record)


class AbstractBufferProcessor(AbstractRecordProcessor, AsyncOnStart, AsyncOnStop, ABC):
    @dataclass
    class Config(AbstractBuffer.Config, AbstractRecordProcessor.Config):
        exception_amount_threshold = 5

    @dataclass
    class Context(AbstractRecordProcessor.Context):
        pass

    def __init__(self,
                 config: Config,
                 context: Context,
                 record_callback: Callable[[list[KafkaRecord]], Awaitable[None]]
                 ) -> None:
        super().__init__(config, context, record_callback)
        self.config = config
        self._buffer = AbstractBuffer(self._handle_buffer, config)
        self._buffer_task: Optional[asyncio.Task] = None

    async def _on_start(self) -> None:
        self._buffer_task = asyncio.create_task(self._buffer.run())

    async def _on_stop(self) -> None:
        self._buffer_task.cancel()

    async def process_record(self, record: ConsumerRecord) -> None:
        await self._buffer.add_to_buffer(record)

    async def _handle_buffer(self, buffer: list[ConsumerRecord]) -> None:
        logger.debug("handle buffer")

        last_record = buffer[-1]
        formed_records = []
        exception_amount = 0
        for record in buffer:
            try:
                pre_proceeded_record = await self._preprocess_record(record)
                formed_records.append(self.form_record(pre_proceeded_record, record))
            except PreProcessRecordError as e:
                # TODO consider add option to skip preprocessing exception and commit anyway
                logger.warning(f"got preprocess error on record {record}, error: {repr(e)}")
                exception_amount += 1
                if exception_amount >= self.config.exception_amount_threshold:
                    # TODO fix; not working properly
                    raise e
        await acall(self._record_callback(formed_records))
        await self.context.consumer.commit(last_record)


class RecordProcessor(AbstractDeserializedRecordProcessor, AbstractSingleRecordProcessor):
    @dataclass
    class Config(AbstractSingleRecordProcessor.Config):
        pass

    @dataclass
    class Context(AbstractDeserializedRecordProcessor.Context):
        pass


class BufferRecordProcessor(AbstractDeserializedRecordProcessor, AbstractBufferProcessor):
    @dataclass
    class Config(AbstractBufferProcessor.Config, AbstractRecordProcessor.Config):
        pass

    @dataclass
    class Context(AbstractDeserializedRecordProcessor.Context):
        pass

    def __init__(self,
                 config: Config,
                 context: Context,
                 record_callback: Callable[[list[KafkaRecord]], Awaitable[None]]
                 ) -> None:
        super().__init__(config, context, record_callback)
